# -*- coding:utf-8 -*-
import logging
from enum import Enum, unique

from restsql.config.database import *
from restsql.config.table import Table
from restsql.config.value import *

__all__ = ['DatabaseType', 'register_database', 'register_table']

logger = logging.getLogger('config')


@unique
class DatabaseType(Enum):
    MYSQL = 'MYSQL'
    POSTGRESQL = 'POSTGRESQL'
    IMPALA = 'IMPALA'
    ELASTICSEARCH = 'ELASTICSEARCH'


def register_database(name, db_type, host,
                      port=None,
                      user=None,
                      password=None,
                      db_name=None,
                      ):
    if name in index4database:
        logger.warning('发现同名数据源%s，将忽略非首次出现的数据源配置', name)
        return
    database = None
    if DatabaseType.MYSQL.value == db_type:
        database = MySQLDatabase(host, port, user, password, db_name)
    elif DatabaseType.POSTGRESQL.value == db_type:
        database = PgDatabase(host, port, user, password, db_name)
    elif DatabaseType.ELASTICSEARCH.value == db_type:
        database = EsDatabase(host)
    elif DatabaseType.IMPALA.value == db_type:
        database = ImpalaDatabase(host, port, db_name)
    else:
        logger.warning('发现未知数据源类型%s，将忽略本次数据源载入', db_type)
    databases.append(database)
    index4database[name] = database


def register_table(name, table_name, fields, belong_to):
    index_name = '{}.{}'.format(belong_to, name)
    if index_name in index4table:
        logger.warning('发现同名表%s，将忽略非首次出现的表配置', index_name)
        return
    table = Table(name, table_name, fields)
    tables.append(table)
    index4table[index_name] = table
    # 存储相应表与数据源中
    database = index4database[belong_to]
    if isinstance(database, Database):
        database.tables.append(table)



# -*- coding:utf-8 -*-

__all__ = ['load', 'database', 'table', 'value']

from restsql.config.load import *
from restsql.config.database import *
from restsql.config.value import *
from restsql.config.table import *



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__all__ = ['RestSqlExceptionBase', 'RunningException', 'UserConfigException', 'JsonFormatException',
           'ProgramConfigException']


class RestSqlExceptionBase(Exception):
    """
    RestSQL库所有异常的基类。
    code 0
    """

    def __init__(self, code, message, *args):
        self.code = code
        self.message = message.format(args)


class RunningException(RestSqlExceptionBase):
    """
    运行时异常；查看日志并处理。
    code [1, 100]
    """
    pass


class UserConfigException(RestSqlExceptionBase):
    """
    用户相关异常；提示用户以矫正。
    code [101, 200]
    """
    pass


class JsonFormatException(UserConfigException):
    """
    json格式异常；提示用户问题。
    code [201, 300]
    """
    pass


class ProgramConfigException(RestSqlExceptionBase):
    """
    检查异常；程序员检查程序。
    code [301, 400]
    """
    pass



# -*- coding:utf-8 -*-

__all__ = ['exception', 'entry', 'engine', 'config']

from restsql.exception import *
from restsql.entry import *
from restsql.engine import *
from restsql.config.load import *
from restsql.config.database import *
from restsql.config.value import *
from restsql.config.table import *




# -*- coding:utf-8 -*-

from enum import Enum, unique

from restsql.exception import ProgramConfigException

__all__ = ['FType', 'Table']


@unique
class FType(Enum):
    INT = 'int'
    BOOL = 'bool'
    NUMBER = 'float'
    STRING = 'string'
    DATETIME = 'datetime'
    TIMEDELTA = 'timedelta'


class Table:

    def __init__(self, name, table_name, fields):
        """
        :param name: 该表对外展示的名。string。
        :param table_name: 表名。string。
        :param fields: 该表字段。map。
        """
        if isinstance(fields, dict):
            self.name = name
            self.table_name = table_name
            for (fname, ftype) in fields.items():
                if FType.INT.value == ftype:
                    fields[fname] = FType.INT
                elif FType.BOOL.value == ftype:
                    fields[fname] = FType.BOOL
                elif FType.NUMBER.value == ftype:
                    fields[fname] = FType.NUMBER
                elif FType.STRING.value == ftype:
                    fields[fname] = FType.STRING
                elif FType.DATETIME.value == ftype:
                    fields[fname] = FType.DATETIME
                elif FType.TIMEDELTA.value == ftype:
                    fields[fname] = FType.TIMEDELTA
                else:
                    raise ProgramConfigException(301, '表字段类型出错，无法识别的类型{}', ftype)
            self.fields = fields
        else:
            raise ProgramConfigException(301, '无法识别的fields类型: {}', type(fields))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import logging
import re

import pandas as pd

from restsql.config.database import Database
from restsql.config.value import *
from restsql.exception import RunningException, UserConfigException

logging.basicConfig()
logger = logging.getLogger("main")
logger.setLevel("DEBUG")


class RestSQL:
    """
    RestSQL服务提供类。
    """

    JOIN_TYPE_MAP = {
        'left_join': 'left',
        'inner_join': 'inner',
        'full_join': 'outer'
    }

    def __init__(self):
        pass

    @staticmethod
    def query(query_dict):
        """
        查询入口。传入指定格式的字典，返回查询结果的dataframe。
        1. 预处理query_dict: 将子select全部拆出。
        2. 从主select开始，到所有子select完成，执行查询并获得一个个dataframe。
        3. 删除子dataframe中非export的字段与不在on的字段
        4. 将子dataframe join进主dataframe。
        5. 根据fields删除多余字段，并执行计算
        :return: dataframe存储的结果
        """
        # Query
        main_df = RestSQL._single_query(query_dict.get('select'))
        if isinstance(main_df, pd.DataFrame):
            if query_dict.get('join'):
                for join in query_dict.get('join'):  # 子select全部拆除单独query
                    df = RestSQL._single_query(join.get('query').get('select'))
                    # 删除子dataframe中非export与on的字段
                    fields = {}  # 旧列名 -> 目标列名
                    if isinstance(df, pd.DataFrame):
                        cur_fields = list(df)  # 当前列名列表
                        for export in join.get('export'):
                            real, alias = export.split('@', 1)
                            fields[real] = alias
                        for k, v in join.get('on').items():
                            fields[v] = v
                        for field in cur_fields:
                            if field not in fields:  # 当前列多余
                                df.drop(field, axis=1, inplace=True)
                            else:  # 当前列存在，需重命名
                                df.rename({field: fields[field]}, axis=1, inplace=True)
                    # Join
                    main_df = main_df.merge(df, left_on=list(join.get('on').keys()),
                                            right_on=list(join.get('on').values()),
                                            how=RestSQL.JOIN_TYPE_MAP[join.get('type')])
        # Sort
        if len(query_dict.get('sort', [])) != 0:
            ascending_sort = []
            for idx, val in enumerate(query_dict.get('sort')):  # 将'-'号设为倒序，无前缀设为正序
                if val.startswith('-'):
                    query_dict['sort'] = val[1:]
                    ascending_sort.append(False)
                else:
                    ascending_sort.append(True)
            main_df.sort_values(by=query_dict.get('sort'), ascending=ascending_sort, inplace=True)
        # Limit
        main_df.drop(labels=range(query_dict.get('limit', 10000), main_df.shape[0]), axis=0, inplace=True)
        # Alias and calculate
        exclude_list = []
        fields = {}
        for field in query_dict.get('fields'):
            real, alias = field.split('@', 1)
            fields[real] = alias
        columns_list = list(main_df)
        for column in columns_list:
            if column in fields.keys():  # 重命名 or exclude
                if fields[column] == 'exclude':  # 添加到exclude列表，最后统一删除。
                    exclude_list.append(column)
                elif fields[column] != column:
                    main_df.rename({column: fields[column]}, axis=1, inplace=True)
            else:  # 删除多余字段
                main_df.drop(column, axis=1, inplace=True)
        for field in fields:
            if ('+' in field) or ('-' in field) or ('*' in field) or ('/' in field):  # 表达式
                try:
                    main_df.eval('{}={}'.format(fields[field], field), inplace=True)
                except Exception as e:
                    logging.exception(e)
                    raise RunningException(1, '计算 %s 时出错', field)
        for column in exclude_list:  # 删除exclude部分
            main_df.drop(column, axis=1, inplace=True)
        return main_df

    @staticmethod
    def _single_query(select_dict):
        select_from = select_dict.get('from')
        if '.' in select_from:
            database_name, table_name = select_from.split('.', 1)
        else:
            raise UserConfigException(102, '查询中无法拆分from字段: 需要[数据源.数据表]的形式')
        table = index4table.get(select_from, None)
        database = index4database.get(database_name, None)
        if table is None:
            raise RunningException(9, '待查询表格不存在于索引中')
        if database is None:
            raise RunningException(10, '待查询表格所在数据源不存在于索引中')
        select_dict['from'] = table.table_name  # 替换为table.table_name，即可能有schema.table或通配符的
        if isinstance(database, Database):
            stmt, params, fields = database.engine.parse(select_dict)
            logger.debug('Engine get the result:\nstmt = %s\nparams = %s\nfields = %s', stmt, params, fields)
            fields = database.list2map(database_name, table_name, fields)
            df = database.query(stmt, params, fields)  # 真正查询
            return df
        else:
            pass

    @staticmethod
    def _extract_var(expression):
        """
        将计算表达式中的变量提取出来
        :param expression:  (a+b)*(c-d)
        :return: [a,b,c,d]
        """
        return re.findall('[^\+,\-,\*,\/,(,)]+', expression)



# -*- coding:utf-8 -*-

__all__ = ['databases', 'index4database', 'tables', 'index4table']

#
# 将数据池提出，防止循环import问题出现
#

# 数据源
databases = []  # Database
# 数据源索引
index4database = {}  # name -> database (string -> Database)

# 数据表
tables = []  # Table
# 数据表索引
index4table = {}  # database_name.table_name -> table (string -> Table)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

setuptools.setup(
    name='restsql',
    version='2.0.6',
    description=(
        'RestSQL库。用json与数据库交互。'
    ),
    url='https://git.code.oa.com/tencent_cloud_mobile_tools/RestSQL',
    long_description='restsql',
    author="venzozhang",
    author_email='venzozhang@tencent.com',
    maintainer='oliverdding',
    maintainer_email='oliverdding@tencent.com',
    license='MIT License',
    packages=['restsql', 'restsql.config'],
    install_requires=[
        'numpy==1.16.6',
        'pandas==0.24.2',
        'pyyaml==5.4.1',
        'elasticsearch==6.8.1',
        'psycopg2-binary==2.8.6',
        'mysqlclient==1.4.6',
        'impyla==0.16.3'
    ],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import abc
import copy
import logging

from restsql.exception import JsonFormatException, ProgramConfigException


class Engine(metaclass=abc.ABCMeta):
    """
    转换引擎。将grafana传入的json查询格式转为数据库直接使用的查询语句，如sql、dsl等等。
    """

    @abc.abstractmethod
    def parse(self, select_dict):
        """
        将QueryObject对象转为查询的语句。
        :param select_dict: QueryObject对象
        :return: 返回三参数元组，查询语句、查询参数、字段
        """
        pass


class SqlEngine(Engine):

    def __init__(self, quote='"'):
        """
        初始化一个sql engine实例。传入参数为System identifier，也就是为了兼容mysql'`'
        :param quote: system identifier
        """
        self.quote = quote

    def _parse_agg(self, agg):
        """
        处理aggregation段。
        :param agg: 属于aggregation的字符串。
        :return: sql格式的字符串
        """
        if isinstance(agg, str):
            agg_field, agg_suffix = agg.split('__', 1)
            if agg_suffix == 'avg':
                return 'AVG({}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote, agg_field,
                                                   self.quote)
            elif agg_suffix == 'sum':
                return 'SUM({}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote, agg_field,
                                                   self.quote)
            elif agg_suffix == 'max':
                return 'MAX({}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote, agg_field,
                                                   self.quote)
            elif agg_suffix == 'min':
                return 'MIN({}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote, agg_field,
                                                   self.quote)
            elif agg_suffix == 'count':
                return 'COUNT({}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote, agg_field,
                                                     self.quote)
            elif agg_suffix == 'count_distinct':
                return 'COUNT(DISTINCT {}t0{}.{}{}{})'.format(self.quote, self.quote, self.quote,
                                                              agg_field, self.quote)
            else:
                raise ProgramConfigException(401, '请联系开发者，parse agg后缀错误。无法识别的后缀{}。', agg_suffix)
        else:
            raise ProgramConfigException(401, '请联系开发者，parse agg参数错误。需要字符串，而不是{}。', type(agg))

    def _parse_filter(self, key, value):
        """
        处理filter段。
        注意，本方法返回元组，第二个为sql语句的变量部分。为什么不像其他的一样直接编码进去呢？参考了peewee，为了防止时间字段缺少引号。
        :param key: filter元素的左值
        :param value: filter元素的右
        :return: 两元素元组: (sql字符串, 变量字符串 or None)
        """
        if isinstance(key, str):
            if '__' in key:
                field, suffix = key.split('__', 1)
                if suffix == 'gt':
                    return '{}t0{}.{}{}{} > %s'.format(self.quote, self.quote, self.quote, field,
                                                       self.quote), (value,)
                elif suffix == 'gte':
                    return '{}t0{}.{}{}{} >= %s'.format(self.quote, self.quote, self.quote, field,
                                                        self.quote), (value,)
                elif suffix == 'lt':
                    return '{}t0{}.{}{}{} < %s'.format(self.quote, self.quote, self.quote, field,
                                                       self.quote), (value,)
                elif suffix == 'lte':
                    return '{}t0{}.{}{}{} <= %s'.format(self.quote, self.quote, self.quote, field,
                                                        self.quote), (value,)
                elif suffix == 'contains':
                    return '{}t0{}.{}{}{} LIKE {}%{}%{}'.format(self.quote, self.quote, self.quote,
                                                                field, self.quote, self.quote,
                                                                value,
                                                                self.quote), None
                elif suffix == 'startswith':
                    return '{}t0{}.{}{}{} LIKE {}{}%{}'.format(self.quote, self.quote, self.quote,
                                                               field, self.quote, self.quote, value,
                                                               self.quote), None
                elif suffix == 'endswith':
                    return '{}t0{}.{}{}{} LIKE {}%{}{}'.format(self.quote, self.quote, self.quote,
                                                               field, self.quote, self.quote, value,
                                                               self.quote), None
                elif suffix == 'range':
                    if isinstance(value, list) and len(value) == 2:
                        return '{}t0{}.{}{}{} BETWEEN %s AND %s'.format(self.quote, self.quote, self.quote,
                                                                        field, self.quote), value
                    else:
                        raise JsonFormatException(3, 'range需要传入长度为2的范围列表。如: [1, 2]。')
                elif suffix == 'in':
                    if isinstance(value, list):
                        template = '%s, {}'
                        in_str = ''
                        for i in range(len(value)):
                            in_str = template.format(in_str)
                        in_str = in_str.rsplit(',', 1)[0]
                        return '{}t0{}.{}{}{} IN ({})'.format(self.quote, self.quote, self.quote,
                                                              field, self.quote, in_str), value
                    else:
                        raise JsonFormatException(3, 'in需要传入元素列表，而不是{}。', type(value))
                else:
                    raise ProgramConfigException(401, '请联系开发者，parse filter后缀错误。无法识别的后缀{}。', suffix)
            else:
                return '{}t0{}.{}{}{} == %s'.format(self.quote, self.quote, self.quote, key,
                                                    self.quote), (value,)
        else:
            raise ProgramConfigException(401, '请联系开发者，parse filter参数错误。需要字符串，而不是{}。', type(key))

    def parse(self, select_dict):
        """
        将QueryObject对象转为sql查询语句。
        确保from字段为table_name，而不是database_name.table_name
        :param select_dict: select_dict部分dict
        :return: 返回三元素元组，第一个是查询语句，第二个是参数列表，第三个是查询结果的列名列表
        """
        if isinstance(select_dict, dict):
            #
            # 拆分
            #
            params_list = []  # 最终sql参数
            fields_list = []  # 最终列表列名
            # select字段
            select_part = []
            for field in list(select_dict.get('fields')):  # select.fields
                select_part.append('{}t0{}.{}{}{}'.format(self.quote, self.quote, self.quote, field, self.quote))
            fields_list.extend(select_dict.get('fields'))
            if select_dict.get('aggregation'):
                for agg in select_dict.get('aggregation'):  # select.aggregation
                    select_part.append(self._parse_agg(agg))
                    fields_list.append(agg)
            # where字段
            where_part = []
            if select_dict.get('filter'):
                for (k, v) in select_dict.get('filter').items():  # select.filter
                    where, params = self._parse_filter(k, v)
                    where_part.append(where)
                    if params is not None:
                        params_list.extend(list(params))
            # group by 字段
            group_by_part = []
            if select_dict.get('group_by'):
                for group_by_field in select_dict.get('group_by'):  # main-group-by
                    group_by_part.append('{}t0{}.{}{}{}'.format(self.quote, self.quote, self.quote, group_by_field, self.quote))
            #
            # 拼凑SQL语句
            #
            query = 'SELECT '
            for select in select_part:
                if select == select_part[-1]:
                    query += '{} '.format(select)
                else:
                    query += '{}, '.format(select)
            select_from = select_dict.get('from')
            select_from = select_from.replace('.', '"."')
            query += 'FROM {}{}{} AS {}t0{} '.format(self.quote, select_from, self.quote, self.quote,
                                                     self.quote)
            if len(where_part) != 0:
                query += 'WHERE ( ( TRUE ) '
            for where in where_part:
                query += 'AND ( {} ) '.format(where)
            if len(where_part) != 0:
                query += ') '
            if len(group_by_part) != 0:
                query += 'GROUP BY '
            for group_by in group_by_part:
                if group_by == group_by_part[-1]:
                    query += '{} '.format(group_by)
                else:
                    query += '{}, '.format(group_by)
            query += 'LIMIT {}'.format(select_dict.get('limit', 10000))
            return query, params_list, fields_list
        else:
            raise ProgramConfigException(400, '请联系开发者，parse object参数错误。请传入query的select部分')


class EsEngine(Engine):
    """
    ElasticSearch查询引擎
    """

    _template = {
        'size': 10000,
        'query': {  # Query context
            'bool': {
                'filter': [  # Filter context
                ],
            },
        },
        '_source': {
            'includes': []
        },
        'aggs': {
            'groupby': {
                'terms': {
                    'script': {
                        'source': ''
                    },
                    'size': 10000,
                },
                'aggs': {}
            },
        },
    }

    def _parse_filter(self, filters):
        must_list = []
        for key, val in filters.items():
            if '__' in key:
                field, suffix = key.split('__', 1)
                if suffix == 'gt':
                    must_list.append({
                        'range': {
                            field: {'gt': val}
                        }
                    })
                elif suffix == 'gte':
                    must_list.append({
                        'range': {
                            field: {'gte': val}
                        }
                    })
                elif suffix == 'lt':
                    must_list.append({
                        'range': {
                            field: {'lt': val}
                        }
                    })
                elif suffix == 'lte':
                    must_list.append({
                        'range': {
                            field: {'lte': val}
                        }
                    })
                elif suffix == 'contains':
                    """"
                    TODO: 本来想用match/match_phrase来进行模糊匹配，但是由于这两种查询由于分词的缘故，现有的
                          分词情况并不能完美的模拟sql中的like，所以暂时采用正则查询。正则查询的效率很低。
                    must_list.append({
                        'match_phrase': {
                            field_name: {
                                'query': value
                            }
                        }
                    })
                    """
                    must_list.append({
                        'wildcard': {field: ''.join(['*', val, '*'])}
                    })
                elif suffix == 'startswith':
                    must_list.append({
                        'prefix': {field: val}
                    })
                elif suffix == 'endswith':
                    must_list.append({
                        'wildcard': {field: ''.join(['*', val])}
                    })
                elif suffix == 'range':
                    if len(val) != 2:
                        raise JsonFormatException(3, 'range需要传入长度为2的范围列表。如: [1, 2]。')
                    must_list.append({
                        'range': {
                            field: {'gte': val[0], 'lte': val[1]}
                        }
                    })
                elif suffix == 'in':
                    if isinstance(val, list):
                        must_list.append({
                            'terms': {field: val}
                        })
                    else:
                        raise JsonFormatException(3, 'in需要传入元素列表，而不是{}。', type(val))
                else:
                    raise ProgramConfigException(401, '请联系开发者，parse filter后缀错误。无法识别的后缀{}。', suffix)
            else:
                must_list.append({
                    'term': {
                        key: val
                    }
                })
        return must_list

    def parse(self, select_dict):
        """
        转为elasticsearch的查询dsl。
        :param select_dict: select_dict部分dict
        :return: 返回三元素元组，第一个是查询语句，第二个是index，第三个是fields
        """
        if isinstance(select_dict, dict):
            fields = []
            dsl = copy.deepcopy(EsEngine._template)
            dsl['size'] = select_dict.get('limit', 10000)
            dsl['aggs']['groupby']['terms']['size'] = select_dict.get('limit', 10000)
            dsl['_source']['includes'].extend([field for field in select_dict.get('fields')])

            if select_dict.get('filter'):
                must_list = self._parse_filter(select_dict.get('filter'))
            else:
                must_list = []
            dsl['query']['bool']['filter'].extend(must_list)
            if not dsl['query']['bool']['filter']:  # 若filter为空直接删除
                del dsl['query']['bool']['filter']
            if not dsl['query']['bool']:  # 若bool为空直接删除
                del dsl['query']['bool']
            if not dsl['query']:  # 若query为空直接删除
                del dsl['query']

            dsl_group_by = ''
            dsl_aggs = dsl['aggs']['groupby']['aggs']
            if select_dict.get('group_by') and len(select_dict.get('group_by')) != 0:  # 只有有group by，agg才有意义
                """
                由于ES 6.x以下版本不支持 composite 语法，所以这里采用script方式来实现group by，用来兼容不同版本ES这部分语法的差异性
                script中source的格式：key:value;key:value
                定义成这个样子是方便后面从查询结果中提取数据
                """
                for field in select_dict.get('group_by', []):
                    dsl_group_by = ''.join(
                        [dsl_group_by, "'", field, "'", " + ':' + ", "doc['", field, "'].value", " + ';' + "])
                dsl_group_by = dsl_group_by[:len(dsl_group_by) - len(" + ';' + ")]  # 去掉结尾的 " + ';' + "
                dsl['aggs']['groupby']['terms']['script']['source'] = dsl_group_by
                # 处理 aggregation
                func_map = {'count': 'value_count', 'sum': 'sum', 'avg': 'avg', 'max': 'max', 'min': 'min',
                            'count_distinct': 'cardinality'}
                for agg in select_dict.get('aggregation', []):
                    field, suffix = agg.split('__', 1)
                    if suffix in func_map:
                        dsl_aggs['{}__{}'.format(field, suffix)] = {
                            func_map[suffix]: {'field': field}}
                    else:
                        raise ProgramConfigException(401, '请联系开发者，parse agg后缀错误。无法识别的后缀{}。', suffix)
            else:
                del dsl['aggs']
            return dsl, select_dict.get('from'), None
        else:
            raise ProgramConfigException(400, '请联系开发者，parse object参数错误。需要dict类实例')



# -*- coding:utf-8 -*-
import abc

import MySQLdb
import impala.dbapi as impyla
import pandas as pd
import psycopg2
from elasticsearch import Elasticsearch

from restsql.config.table import FType
from restsql.config.value import *
from restsql.engine import SqlEngine, EsEngine
from restsql.exception import ProgramConfigException, RunningException

__all__ = ['Database', 'PgDatabase', 'MySQLDatabase', 'ImpalaDatabase', 'EsDatabase']


_BOOL_MAP = {
    'f': False,
    'F': False,
    'fal': False,
    'Fal': False,
    'false': False,
    'False': False,
    'FALSE': False,
    False: False,
    't': True,
    'T': True,
    'tr': True,
    'Tr': True,
    'true': True,
    'True': True,
    'TRUE': True,
    True: True,
}


class Database(metaclass=abc.ABCMeta):
    """
    数据源抽象类。存储与数据库连接相关信息。
    提供查询数据入口。
    """

    def __init__(self):
        self.tables = []
        self.conn = None
        self.engine = None

    def list2map(self, database_name, table_name, fields):
        """
        判断查询字段是否存在表中，并将字段列表转为map(field_name -> field_type)
        self.tables为属于本数据源的数据表。
        :param database_name: 目标数据源的name
        :param table_name: 目标数据表的name
        :param fields: 待检测字段
        :return: 若传入fields为list则返回转换后的map；否则返回待查表所有fields
        """
        index_name = '{}.{}'.format(database_name, table_name)
        if index_name not in index4table:  # 系统中不存在该表
            raise RunningException(4, '系统中不存在待查询的表')
        table = index4table[index_name]
        if table not in self.tables:  # 数据源中不存在该表
            raise RunningException(5, '目标数据源中不存在待查询的表')
        if isinstance(fields, list):
            field_map = {}
            for field in fields:
                if '__' in field:
                    field_temp = field.split('__', 1)[0]
                    if field_temp not in table.fields:  # 字段不存在
                        raise RunningException(6, '待查询表中不存在待查询字段')
                    field_map[field] = table.fields[field_temp]
                else:
                    if field not in table.fields:  # 字段不存在
                        raise RunningException(6, '待查询表中不存在待查询字段')
                    field_map[field] = table.fields[field]
            return field_map
        else:
            return table.fields

    @abc.abstractmethod
    def query(self, stmt, params, fields):
        """
        执行查询，并返回只包含fields的dataframe结构。
        self.conn为数据库连接，根据不同的库提供的api调用对应方法完成需求。
        应为静态方法，即每个数据源类型的查询方式都交由程序员实现。
        :param stmt: 查询语句。
        :param params: 查询参数。
        :param fields: 查询字段。
        """
        pass


class _SqlDatabase(Database):
    """
    Sql相关类的公共父类，抽取相同的sql逻辑于此。
    """

    def query(self, stmt, params, fields):
        if isinstance(fields, dict):
            # query
            cur = self.conn.cursor()
            cur.execute(stmt, params)
            # fetch
            result = list(cur.fetchall())
            cur.close()
            self.conn.rollback()
            # parse
            fnames = list(fields.keys())
            df = pd.DataFrame.from_records(result, columns=fnames)
            # type
            for fname in fnames:
                if fields[fname] == FType.INT:
                    df[fname] = df[fname].astype('int')
                elif fields[fname] == FType.BOOL:
                    df[fname] = df[fname].map(_BOOL_MAP)
                elif fields[fname] == FType.NUMBER:
                    df[fname] = df[fname].apply('float')
                elif fields[fname] == FType.DATETIME:
                    df[fname] = df[fname].apply(pd.to_datetime)
                elif fields[fname] == FType.TIMEDELTA:
                    df[fname] = df[fname].apply(pd.to_datetime)
            return df
        else:
            raise ProgramConfigException(301, '无法识别的fields类型: {}', type(fields))


class PgDatabase(_SqlDatabase):

    def __init__(self, host, port, user, password, db_name):
        super().__init__()
        self.engine = SqlEngine()
        self.conn = psycopg2.connect(
            database=db_name,
            user=user,
            password=password,
            host=host,
            port=int(port),
        )


class MySQLDatabase(_SqlDatabase):

    def __init__(self, host, port, user, password, db_name):
        super().__init__()
        self.engine = SqlEngine('`')
        self.conn = MySQLdb.connect(
            db=db_name,
            user=user,
            passwd=password,
            host=host,
            port=int(port)
        )


class ImpalaDatabase(_SqlDatabase):

    def __init__(self, host, port, db_name):
        super().__init__()
        self.conn = impyla.connect(
            host=host,
            port=port,
            database=db_name,
        )
        self.engine = SqlEngine()


class EsDatabase(Database):

    def __init__(self, host):
        super().__init__()
        self.conn = Elasticsearch(host)
        self.engine = EsEngine()

    def query(self, stmt, params, fields):
        """
        执行查询并返回结果dataframe。结果dataframe中可能包含无关字段。 TODO: 实现这部分
        :param stmt: 查询语句
        :param params: 查询index
        :param fields: 字段->类型的dict
        """
        raw_result = self.conn.search(index=params, body=stmt)
        df = None
        if 'aggs' in raw_result or 'aggregations' in raw_result:  # dsl中有aggs.groupby.aggs
            if raw_result.get('aggregations'):
                if not raw_result.get('aggregations').get('groupby').get('buckets'):
                    return pd.DataFrame(columns=[])
                df = pd.DataFrame(raw_result['aggregations']['groupby']['buckets'])
            else:
                if not raw_result.get('aggregations').get('groupby').get('buckets'):
                    return pd.DataFrame(columns=[])
                df = pd.DataFrame(raw_result['aggs']['groupby']['buckets'])
            df.drop('doc_count', axis=1, inplace=True)
            # Dealing with others but key
            columns = list(df)
            for column in columns:
                if column != 'key':
                    df[column] = df[column].map(lambda x: x['value'])
            # Dealing with key
            example_val = df['key'].iat[0]
            col_count = len(example_val.split(';'))
            for i in range(col_count):
                col_name = example_val.split(';')[i].split(':', 1)[0]
                df[col_name] = df['key'].map(lambda x: x.split(';')[i].split(':', 1)[1])
            df.drop('key', axis=1, inplace=True)
        elif 'hits' in raw_result and 'hits' in raw_result['hits']:
            df = pd.concat(map(pd.DataFrame.from_dict, raw_result['hits']['hits']), axis=1)['_source'].T
            df.reset_index(drop=True, inplace=True)  # 重置索引
        else:
            raise RunningException(1, 'es查询失败: 未在结果中看到aggs、hits数据部分。请查看日志')
        # 字段类型转换
        columns = list(df)
        for column in columns:
            column_without_suffix = column
            if '__' in column:
                column_without_suffix = column.split('__', 1)[0]
            if column_without_suffix in fields:  # 该字段存在 -> 将该字段的字段格式转为基本格式类型
                if fields[column_without_suffix] == FType.INT:
                    df[column] = df[column].astype('int')
                elif fields[column_without_suffix] == FType.NUMBER:
                    df[column] = df[column].apply('float')
                elif fields[column_without_suffix] == FType.BOOL:
                    df[column] = df[column].map(_BOOL_MAP)
                elif fields[column_without_suffix] == FType.DATETIME:
                    df[column] = df[column].apply(pd.to_datetime)
                elif fields[column_without_suffix] == FType.TIMEDELTA:
                    df[column] = df[column].apply(pd.to_datetime)
            else:
                raise RunningException(8, 'ES查询结果中未知字段: {}', column)
        return df
